from __future__ import annotations

import chik_rs.sized_bytes

bytes4 = chik_rs.sized_bytes.bytes4
bytes8 = chik_rs.sized_bytes.bytes8
bytes32 = chik_rs.sized_bytes.bytes32
bytes48 = chik_rs.sized_bytes.bytes48
bytes96 = chik_rs.sized_bytes.bytes96
bytes100 = chik_rs.sized_bytes.bytes100
bytes480 = chik_rs.sized_bytes.bytes480
