from __future__ import annotations

SPACEBUCKS = {
    "asset_id": "a628c1c2c6fcb74d53746157e438e108eab5c0bb3e5c80ff9b1910b3e4832913",
    "name": "Spacebucks",
    "symbol": "SBX",
}

MARMOT = {
    "asset_id": "8ebf855de6eb146db5602f0456d2f0cbe750d57f821b6f91a8592ee9f1d4cf31",
    "name": "Marmot",
    "symbol": "MRMT",
}

STABLY_USDS = {
    "asset_id": "6d95dae356e32a71db5ddcb42224754a02524c615c5fc35f568c2af04774e589",
    "name": "Stably USD Classic",
    "symbol": "USDSC",
}

CHIK_HOLIDAY_TOKEN = {
    "asset_id": "509deafe3cd8bbfbb9ccce1d930e3d7b57b40c964fa33379b18d628175eb7a8f",
    "name": "Chik Holiday 2023 Token",
    "symbol": "CK23",
}

BASE_WARPED_USDC = {
    "asset_id": "fa4a180ac326e67ea289b869e3448256f6af05721f7cf934cb9901baa6b7a99d",
    "name": "Base Warped USDC ",
    "symbol": "wUSDC.b",
}

BASE_WARPED_MILLIETH = {
    "asset_id": "f322a205c034fe28681829fa5a2e483ac421f0952eb1292945c8db06e0a471a6",
    "name": "Base Warped milliETH",
    "symbol": "wmilliETH.b",
}

BASE_WARPED_EURC = {
    "asset_id": "31ffd54c5b38bb33352dc0be0ebf9cf4f29a6329156dd4811d6f4254f85c8200",
    "name": "Base Warped EURC",
    "symbol": "wEURC.b",
}


ETH_WARPED_USDC = {
    "asset_id": "bbb51b246fbec1da1305be31dcf17151ccd0b8231a1ec306d7ce9f5b8c742b9e",
    "name": "Ethereum Warped USDC",
    "symbol": "wUSDC",
}

ETH_WARPED_MILLIETH = {
    "asset_id": "4cb15a8ecc85068fb1f98c09a5e489d1ad61b2af79690ce00f9fc4803c8b597f",
    "name": "Ethereum Warped milliETH",
    "symbol": "wmilliETH",
}

ETH_WARPED_USDT = {
    "asset_id": "634f9f0de1a6c39a2189948b8e61b6852fbf774f73b0e36e143e841c49a0798c",
    "name": "Ethereum Warped USDT",
    "symbol": "wUSDT",
}

DEFAULT_CATS = {
    SPACEBUCKS["asset_id"]: SPACEBUCKS,
    MARMOT["asset_id"]: MARMOT,
    STABLY_USDS["asset_id"]: STABLY_USDS,
    CHIK_HOLIDAY_TOKEN["asset_id"]: CHIK_HOLIDAY_TOKEN,
    BASE_WARPED_USDC["asset_id"]: BASE_WARPED_USDC,
    BASE_WARPED_MILLIETH["asset_id"]: BASE_WARPED_MILLIETH,
    BASE_WARPED_EURC["asset_id"]: BASE_WARPED_EURC,
    ETH_WARPED_USDC["asset_id"]: ETH_WARPED_USDC,
    ETH_WARPED_MILLIETH["asset_id"]: ETH_WARPED_MILLIETH,
    ETH_WARPED_USDT["asset_id"]: ETH_WARPED_USDT,
}
