from __future__ import annotations

import chik_rs.sized_ints

int8 = chik_rs.sized_ints.int8
uint8 = chik_rs.sized_ints.uint8

int16 = chik_rs.sized_ints.int16
uint16 = chik_rs.sized_ints.uint16

int32 = chik_rs.sized_ints.int32
uint32 = chik_rs.sized_ints.uint32

int64 = chik_rs.sized_ints.int64
uint64 = chik_rs.sized_ints.uint64

uint128 = chik_rs.sized_ints.uint128

int512 = chik_rs.sized_ints.int512
